
/**
 * @imports
 */
import {Reflex} from '@web-native-js/chtml';
import _isArray from '@web-native-js/commons/js/isArray.js';
import _isEmpty from '@web-native-js/commons/js/isEmpty.js';
import _isFunction from '@web-native-js/commons/js/isFunction.js';
import _arrFirst from '@web-native-js/commons/arr/first.js';
import _arrLast from '@web-native-js/commons/arr/last.js';
import _following from '@web-native-js/commons/arr/following.js';
import _preceding from '@web-native-js/commons/arr/preceding.js';
import _each from '@web-native-js/commons/obj/each.js';
import CollectionBase from './CollectionBase.js';

/**
 * ---------------------------
 * The List class
 * ---------------------------
 */
			
export default class extends CollectionBase {

	/**
	 * Constructs a new List instance.
	 * Sub-views may also be listed.
	 *
	 * @param array|object			items
	 * @param object				params
	 *
	 * @return void
	 */
	constructor(items = {}, params = {}) {
		super(items, params);
		var itemActiveStatePath = this.params.itemsOffset + '..active';
		Reflex.observe(this, itemActiveStatePath, (activeState, _activeState, e) => {
			if (activeState === 'activating') {
				this.getItemsArray().forEach(item => {
					if (item.active === true) {
						item.setActiveState(false);
					}
				});
			} else if (activeState === true) {
				// Set lone activeChild
				return Reflex.set(this, {
					activeChild: e.originatingTarget,
					hasActiveChild: true,
				}, true/*returnEvent*/);
			} else if (activeState === false) {
				// Unset lone activeChild
				return Reflex.set(this, {
					activeChild: undefined,
					hasActiveChild: false,
				}, true/*returnEvent*/);
			}
		});
	}

	/**
	 * Advances the list's selectedness to the
	 * first item.
	 *
	 * @return void}Event
	 */
	selectStart() {
		var first, items = this.getItemsArray();
		if (!_isEmpty(items) && (first = _arrFirst(items))) {
			return first.setActiveState(true);
		}
	}

	/**
	 * Advances the list's selectedness to the
	 * last item.
	 *
	 * @return void}Event
	 */
	selectEnd() {
		var last, items = this.getItemsArray();
		if (!_isEmpty(items) && (last = _arrLast(items))) {
			return last.setActiveState(true);
		}
	}
	
	/**
	 * Advances the list's selectedness to the
	 * item preceding the current activeChild.
	 * Selects the last item if loopable and no activeChild.
	 *
	 * @param bool|function	loop
	 *
	 * @return void}Event
	 */
	selectPrev(loop = false) {
		var preceding, items = this.getItemsArray();
		if (_isEmpty(items)) {
			if (_isFunction(loop)) {
				loop();
			}
			return;
		}
		if (this.activeChild) {
			preceding = _preceding(items, this.activeChild, false/*length*/, loop);
		} else if (loop && (!_isFunction(loop) || loop(0))) {
			preceding = _arrLast(items);
		}
		if (preceding) {
			return preceding.setActiveState(true);
		}
	}

	/**
	 * Advances the list's selectedness to the
	 * item following the current activeChild.
	 * Selects the first item if loopable and no activeChild.
	 *
	 * @param bool|function	loop
	 *
	 * @return void}Event
	 */
	selectNext(loop = false) {
		var following, items = this.getItemsArray();
		if (_isEmpty(items)) {
			if (_isFunction(loop)) {
				loop();
			}
			return;
		}
		if (this.activeChild) {
			following = _following(items, this.activeChild, false/*length*/, loop);
		} else if (loop && (!_isFunction(loop) || loop(0))) {
			following = _arrFirst(items);
		}
		if (following) {
			return following.setActiveState(true);
		}
	}

	/**
	 * Advances the list's selectedness to a random item.
	 *
	 * @return void}Event
	 */
	selectRand() {
	}

	/**
	 * Advances the list's selectedness to a item.
	 *
	 * @return object
	 */
	filter() {
	}

	/**
	 * Advances the list's selectedness to a item.
	 *
	 * @return object
	 */
	sort() {
	}
};