
/**
 * @imports
 */
import _sort from '@web-native-js/commons/arr/sort.js';
import _difference from '@web-native-js/commons/arr/difference.js';
import _isArray from '@web-native-js/commons/js/isArray.js';
import _isObject from '@web-native-js/commons/js/isObject.js';
import _isString from '@web-native-js/commons/js/isString.js';
import _isEmpty from '@web-native-js/commons/js/isEmpty.js';
import _copy from '@web-native-js/commons/obj/copy.js';
import _with from '@web-native-js/commons/obj/with.js';
import {Reflex} from '@web-native-js/chtml';
import globalParams from '../../params.js';

/**
 * ---------------------------
 * The Router class
 * ---------------------------
 *
 * This class provides additional functionality over
 * the native document.history object and the "onpopstate" event.
 */
			
export default class Router {

	/**
	 * Constructs a new Router instance. Typically,
	 * only one instance would be needed app-wide. So an should
	 * be used as a singleton.
	 *
	 * @return void
	 */
	constructor() {
		// -----------------------
		// Helpers
		var getProps = (urlObj, state = null) => {
			urlObj = _copy(urlObj, urlProperties, false/*withSymbols*/);
			return state ? _with(urlObj, 'state', state) : urlObj;
		};
		var _strictEven = (a, b) => {
			if (_isObject(a) && _isObject(b)) {
				return _strictEven(Object.keys(a), Object.keys(b)) 
				&& _strictEven(Object.values(a), Object.values(b));
			}
			if (_isArray(a) && _isArray(b)) {
				return a.length === b.length 
				&& a.reduce((recieved, item, i) => recieved && item === b[i], true);
			}
			return a === b;
		};
		// -----------------------
		// Setting the "href" properties must also publish
		// all the other properties in urlProperties
		Reflex.trap(this, (e, recieved, next) => {
			if (e.query === 'href' && _difference(urlProperties, e.related).length) {
				var urlObj = Router.parseUrl(e.value);
				Reflex.set(this, urlObj);
				return false;
			}
			return next();
		}, {type:'set'});
		// -----------------------
		// When any one of these properties change,
		// the others are automatically derived
		Reflex.observe(this, ['search', 'searchmap', 'pathname', 'pathmap', 'pathsplit', 'href'], (vals, _vals, e) => {
			var [search, searchmap, pathname, pathmap, pathsplit, href] = vals;
			var [_search, _searchmap, _pathname, _pathmap, _pathsplit, _href] = _vals;
			if (e.originatingCreated && e.originatingCreated.indexOf('searchmap') > -1) {
				return;
			}
			var urlObj = {};
			// ----------
			var isSearchmapSame = !e.originatingFields.filter(f => f.startsWith('searchmap.')).length && _strictEven(searchmap, _searchmap);
			if (search === _search && !isSearchmapSame) {
				// "searchmap" was updated. So we update "search"
				urlObj.search = Router.toSearch(searchmap);
				if (urlObj.search === this.search) {
					delete urlObj.search;
				}
			}
			if (search !== _search && isSearchmapSame) {
				// "search" was updated. So we update "searchmap"
				urlObj.searchmap = Router.toSearchmap(search);
				if (_strictEven(urlObj.searchmap, this.searchmap)) {
					delete urlObj.searchmap;
				}
			}
			// ----------
			var isPathmapSame = !e.originatingFields.filter(f => f.startsWith('pathmap.')).length && _strictEven(pathmap, _pathmap);
			var isPathsplitSame = !e.originatingFields.filter(f => f.startsWith('pathsplit.')).length && _strictEven(pathsplit, _pathsplit);
			if (pathname === _pathname && (!isPathmapSame || !isPathsplitSame)) {
				// We update "pathname" from the new "pathmap"/"pathsplit"
				urlObj.pathname = Router.toPathname(!isPathsplitSame ? pathsplit : pathmap, this.pathname/*referenceUrl*/);
				if (urlObj.pathname === this.pathname) {
					delete urlObj.pathname;
				}
			}
			if (isPathmapSame && (pathname !== _pathname || !isPathsplitSame)) {
				// We update "pathmap" from the new "pathname"/"pathsplit"
				urlObj.pathmap = Router.toPathmap(pathname !== _pathname ? pathname : pathsplit);
				if (_strictEven(urlObj.pathmap, this.pathmap)) {
					delete urlObj.pathmap;
				}
			}
			if (isPathsplitSame && (pathname !== _pathname || !isPathmapSame)) {
				// We update "pathsplit" from the new "pathname"/"pathmap"
				urlObj.pathsplit = Router.toPathsplit(pathname !== _pathname ? pathname : pathmap, this.pathname/*referenceUrl*/);
				if (_strictEven(urlObj.pathsplit, this.pathsplit)) {
					delete urlObj.pathsplit;
				}
			}
			// ----------
			if (href === _href && !_isEmpty(urlObj)) {
				// We update "href" from the new component values
				urlObj.href = this.origin;
				urlObj.href += urlObj.pathname/*if pathmap or pathsplit was the change*/ || pathname/*whether or not pathname was the change*/;
				urlObj.href += urlObj.search/*if searchmap was the change*/ || search/*whether or not search was the change*/ || '';
				if (urlObj.href === this.href) {
					delete urlObj.href;
				}
			} else if (_isEmpty(urlObj) && href !== _href) {
				// We update component values from the new "href"
				urlObj = Router.parseUrl(href);
			}
			if (!_isEmpty(urlObj)) {
				return Reflex.set(this, urlObj, true/*returnEvent*/);
			}
		}, {observeDown:true/*for pathmap/pathsplit/searchmap updates*/});
		// -----------------------
		// Syndicate changes to
		// the browser;s location bar
		Reflex.observe(this, 'href', href => {
			if (href === document.location.href) {
				window.history.replaceState(this.state, '', href);
			} else {
				window.history.pushState(this.state, '', href);
			}
		});
		// -----------------------
		// This event is triggered by
		// either the browser back button,
		// the window.history.back(),
		// the window.history.forward(),
		// or the window.history.go() action.
		window.addEventListener('popstate', e => {
			// Needed to alow document.location
			// to update to window.location
			setTimeout(() => {
				Reflex.set(this, getProps(document.location, window.history.state));
			}, 0);
		});
		// -----------------------
		// Capture all link-clicks
		// and fire to this router.
		window.addEventListener('click', e => {
			// TODO: remove without issues
			e.preventDefault();
			var anchor, href, target;
			if ((anchor = e.target.closest('a')) 
			&& (href = anchor.href)) {
				var e2 = this.route(href, null, anchor.getAttribute('target'));
				if (e2 && e2.defaultPrevented) {
					e.preventDefault();
				}
			}
		});
		// -----------------------
		// Startup properties
		if (!this.href) {
			// Startup route
			Reflex.set(this, getProps(document.location));
		}
		if (!this.state && window.history) {
			// Startup state
			Reflex.set(this, 'state', window.history.state);
		}
	}

	/**
	 * Sets new URL and state data and fires the change.
	 *
	 * For the benefit of calling observers and accessing listeners' disposition,
	 * this function should be used instead of window.history.pushState() and window.history.replaceState().
	 *
	 * @param string 	href
	 * @param object 	request
	 * @param string 	target
	 *
	 * @return UserEvent
	 */
	route(href, request = null, target = null) {
		var urlObj = {href};
		if (arguments.length > 1) {
			urlObj.request = request;
		}
		if (arguments.length > 2) {
			urlObj.target = target;
		}
		// -----------------
		return Reflex.set(this, urlObj, true/*returnEvent*/);
	}
	
	/**
	 * Updates the current url with new state data.
	 *
	 * For the benefit of calling observers and accessing listeners' disposition,
	 * this function should be used instead of window.history.replaceState().
	 *
	 * @param object 	state
	 *
	 * @return UserEvent
	 */
	update(state) {
		window.history.replaceState(state, '', document.location);
		return Reflex.set(this, 'state', state, true/*returnEvent*/);
	}

	/**
	 * Forwards to a visited route using window.history.forward()
	 *
	 * This function eventually calls observers via "popstate" event bound in the constructor.
	 * But this won't return a Promise as Router.route() would.
	 *
	 * @return void
	 */
	forward() {
		window.history.forward();
	}

	/**
	 * Returns to a previous route using window.history.back()
	 *
	 * This function eventually calls observers via "popstate" event bound in the constructor.
	 * But this won't return a Promise as Router.route() would.
	 *
	 * @return void
	 */
	back() {
		window.history.back();
	}

	/**
	 * Goes to a specific history entry using window.history.go()
	 *
	 * This function eventually calls observers via "popstate" event bound in the constructor.
	 * But this won't return a Promise as Router.route() would.
	 *
	 * @param int key
	 *
	 * @return void
	 */
	go(key) {
		window.history.go(key);
	}

	/**
	 * Tells if the given URL matches the router's current URL
	 *
	 * @param string			href
	 *
	 * @return bool
	 */
	matches(href) {
		var urlObj = Router.parseUrl(href);
		return (this.href + '/').startsWith(!urlObj.href.endsWith('/') ? urlObj.href + '/' : urlObj.href);
	}

	/**
	 * Parses an URL and returns its properties
	 *
	 * @param string			href
	 *
	 * @return object
	 */
	static parseUrl(href) {
		var a = document.createElement('a');
		a.href = href;
		return urlProperties.reduce((obj, prop) => _with(obj, prop, a[prop]), {});
	}

	/**
	 * Parses the input search string into a named map
	 *
	 * @param string			search
	 *
	 * @return object
	 */
	static toSearchmap(search) {
		var queryArr = (search.startsWith('?') ? search.substr(1) : search)
			.split('&').filter(str => str).map(str => str.split('=').map(str => str.trim()));
		return queryArr.reduce((recieved, q) => _with(recieved, q[0], q[1]), {});
	}

	/**
	 * Stringifies the input searchmap to search string.
	 *
	 * @param object			searchmap
	 *
	 * @return string
	 */
	static toSearch(searchmap) {
		return Object.keys(searchmap).length 
			? '?' + Object.keys(searchmap).map(k => k + '=' + searchmap[k]).join('&')
			: '';
	}

	/**
	 * Parses the input path and returns its parts named
	 *
	 * @param string|array			pathnameOrPathsplit
	 *
	 * @return object
	 */
	static toPathmap(pathnameOrPathsplit) {
		var pathArr = _isString(pathnameOrPathsplit) ? pathnameOrPathsplit.split('/').filter(k => k) : pathnameOrPathsplit;
		var pathStr = _isString(pathnameOrPathsplit) ? pathnameOrPathsplit : '/' + pathArr.join('/') + '/';
		var pathnaming = _sort(Object.keys(globalParams.pathnaming), 'desc').reduce((_pathnames, _path) => {
			return _pathnames || ((pathStr + '/').startsWith(_path === '/' ? _path : '/' + _path.split('/').filter(k => k).join('/') + '/') ? globalParams.pathnaming[_path] : null);
		}, null);
		return !pathnaming ? {} : pathArr.reduce((obj, pathItem, i) => pathnaming[i] ? _with(obj, pathnaming[i], pathItem) : obj, {});
	}

	/**
	 * Parses the input path and returns its parts unnamed
	 *
	 * @param string|object			pathnameOrPathmap
	 * @param string				referenceUrl
	 *
	 * @return array
	 */
	static toPathsplit(pathnameOrPathmap, referenceUrl = null) {
		if (_isString(pathnameOrPathmap)) {
			return pathnameOrPathmap.split('/').filter(k => k);
		}
		if (!referenceUrl) {
			throw new Error('A "referenceUrl" must be given to properly determine a path-naming scheme.');
		}
		var pathnaming = _sort(Object.keys(globalParams.pathnaming), 'desc').reduce((_pathnames, _path) => {
			return _pathnames || ((referenceUrl + '/').startsWith(_path === '/' ? _path : '/' + _path.split('/').filter(k => k).join('/') + '/') ? globalParams.pathnaming[_path] : null);
		}, null);
		if (_difference(Object.keys(pathnameOrPathmap), pathnaming).length) {
			throw new Error('The given pathmap contains keys (' + Object.keys(pathnameOrPathmap).join(', ') + ') not recognized by the implied path-naming scheme (' + pathnaming.join(', ') + ')');
		}
		return !pathnaming ? [] : pathnaming.map(name => pathnameOrPathmap[name]).filter(a => a);
	}

	/**
	 * Stringifies the input pathmap or pathsplit to a string
	 *
	 * @param object|array			pathmapOrPathsplit
	 * @param string				referenceUrl
	 *
	 * @return string
	 */
	static toPathname(pathmapOrPathsplit, referenceUrl = null) {
		if (_isObject(pathmapOrPathsplit)) {
			if (!referenceUrl) {
				throw new Error('A "referenceUrl" must be given to properly determine a path-naming scheme.');
			}
			pathmapOrPathsplit = Router.toPathsplit(pathmapOrPathsplit, referenceUrl);
		}
		return '/' + pathmapOrPathsplit.join('/') + '/';
	}
};

/**
 * These are standard
 * and shouldnt'/can't be modified
 *
 * @array
 */
const urlProperties = [
	'hash',
	'host',
	'hostname',
	'href',
	'origin',
	'pathname',
	'port',
	'protocol',
	'search',
];