
/**
 * @imports
 */
import _ListenerBase from '@web-native-js/reflex/src/internal/ListenerBase.js';
import _toTitle from '@web-native-js/commons/str/toTitle.js';
import _arrFrom from '@web-native-js/commons/arr/from.js';
import _find from '@web-native-js/commons/obj/find.js';
import _isString from '@web-native-js/commons/js/isString.js';
import _isFunction from '@web-native-js/commons/js/isFunction.js';
import CustomEvents from './CustomEvents.js';
import UserEvent from './UserEvent.js';
import Listener from './Listener.js';

/**
 * ---------------------------
 * The ListenerBase class
 * ---------------------------
 */				
export default class ListenerBase extends _ListenerBase {

	/**
	 * Constructs a new observable and returns a proxy wrapper.
	 *
	 * @param HTMLElement			el
	 *
	 * @return this
	 */
	constructor(el) {
		super();
		this.el = el;
		this.handlersList = {};
		this.hammertime = new Hammer.Manager(this.el);
	}

	/**
	 * Sets up an event type when its first listener is added.
	 *
	 * @param string TYPE 
	 * @param object params 
	 *
	 * @return void
	 */
	_setup(TYPE, params) {
		// We register an event only once.
		if (this.handlersList[TYPE]) {
			return;
		}
		if (CustomEvents[TYPE]) {
			if (_isString(CustomEvents[TYPE])) {
				// >> LINK THE ALIAS EVENT TO THIS FIREBASE
				this.handlersList[TYPE] = e => this.fire(new UserEvent(this.el, {type:TYPE, e}));
				this.addFireable(new Listener(this.handlersList[TYPE], CustomEvents[TYPE]));
			} else if (_isFunction(CustomEvents[TYPE].setup)) {
				// >> LINK THE CUSTOM EVENT TO THIS FIREBASE
				this.handlersList[TYPE] = true;
				CustomEvents[TYPE].setup(this.el, TYPE, e => {
					if (!_isFunction(CustomEvents[TYPE].validate) 
					|| CustomEvents[TYPE].validate(e, params)) {
						this.fire(new UserEvent(this.el, {type:TYPE, e}))
					}
				}, this.hammertime);
			} else {
				throw new Error('The "' + TYPE + '" event hook must have a "setup" function!');
			}
		} else if (recognizeGesture(TYPE.split('+')[0])) {
			// Lets work as if if always a list
			var recognizers = TYPE.split('+').map(gestureName => {
				var mainGestureName = recognizeGesture(gestureName);
				var recognizer = this.hammertime.get(mainGestureName);
				if (!recognizer) {
					recognizer = new Hammer[_toTitle(mainGestureName)];
					this.hammertime.add(recognizer);
				}
				return recognizer;
			});
			// From right to left, recognizeWith all others ahead
			recognizers.forEach((recognizer, i) => {
				recognizer.recognizeWith(recognizers.slice(i + 1));
			});
			// >> LINK THE HAMMER EVENT TO THIS FIREBASE
			this.handlersList[TYPE] = e => this.fire(new UserEvent(this.el, {type:TYPE, e}));
			this.hammertime.on(TYPE.split('+').join(' '), this.handlersList[TYPE]);
		} else {
			// >> LINK THE DOM EVENT TO THIS FIREBASE
			this.handlersList[TYPE] = e => this.fire(new UserEvent(this.el, {type:TYPE, e}));
			this.el.addEventListener(TYPE, this.handlersList[TYPE]);
		}
	}

	/**
	 * Tears down an event type when its last listener is removed.
	 *
	 * @param string TYPE 
	 *
	 * @return void
	 */
	_teardown(TYPE) {
		// We register an event only once.
		if (!this.handlersList[TYPE]) {
			return;
		}
		if (CustomEvents[TYPE]) {
			if (_isString(CustomEvents[TYPE])) {
				// Level1 custom event
				removeListener(this.el, CustomEvents[TYPE], this.handlersList[TYPE]);
			} else if (_isFunction(CustomEvents[TYPE].teardown)) {
				// Level2 custom event. We supply the this.unobserve method incase it'll be needed.
				CustomEvents[TYPE].teardown(this.el, TYPE, this.hammertime);
			} else {
				throw new Error('The "' + TYPE + '" event hook must have a "teardown" function!');
			}
		} else if (recognizeGesture(TYPE.split('+')[0])) {
			this.hammertime.off(TYPE.split('+').join(' '), this.handlersList[TYPE]);
		} else {
			// Native event
			this.el.removeEventListener(TYPE, this.handlersList[TYPE]);
		}
		delete this.handlersList[TYPE];
	}
	
	/**
	 * @see super.addFireable()
	 *
	 * Captures the params being observed and sets up their hooks
	 * when their first listener just gets added.
	 */
	addFireable(fireable) {
		super.addFireable(fireable);
		this._setup(fireable.params.type, fireable.params);
		return fireable;
	}

	/**
	 * @inheritdoc
	 */
	static createForTarget(object) {
		return super.createForTarget(object, ListenerBase);
	}
};

/**
 * Finds the recognizer for a gestureName.
 *
 * @param string	gestureName
 *
 * @return string
 */
const recognizeGesture = function(gestureName) {
	return _find(gestureIndex, list => list.includes(gestureName), false/*deep*/);
};

/**
 * Gesture reference.
 */
const gestureIndex = {
	press: 	['press', 'pressup',], 
	rotate:	['rotate', 'rotatestart', 'rotatemove', 'rotateend', 'rotatecancel',],
	pinch: 	['pinch', 'pinchstart', 'pinchmove', 'pinchend', 'pinchcancel', 'pinchin', 'pinchout',], 
	pan: 	['pan', 'panstart', 'panmove', 'panend', 'pancancel', 'panleft', 'panright', 'panup', 'pandown',],
	swipe: 	['swipe', 'swipeleft', 'swiperight', 'swipeup', 'swipedown',],
	tap: 	['tap',],
};
