
/**
 * @imports
 */
import _isObject from '@web-native-js/commons/js/isObject.js';
import _any from '@web-native-js/commons/arr/any.js';
import _copy from '@web-native-js/commons/obj/copy.js';

/**
 * Gets the element's width/height values or left/top offsets or both.
 *
 * A context can be specified from which to resolve left/top distances.
 * document origins are used by default.
 *
 * This function can calculate the rect of 3 different types of object:
 * - DOM element: offsets are calculated from el.getBoundingClientRect() and resolved relative to the specified offsetOrigin.
 *		width and height are calculated from el.getBoundingClientRect().
 * - Event object: offsets are calculated as the event's (client|offset|page) x/y, depending on the specified offsetOrigin.
 *		width and height are always 0, 0.
 * - Window object: offsets are calculated as the current left/top scroll, as determined by the value of offsetOrigin.
 *		width and height are always the window's inner width/height values.
 *
 * @param DOMElement|Event|window 	el
 * @param bool	 					size
 * @param DOMElement|window	 		offsetOrigin
 *
 * @return object
 */
const rect = function(el, size = true, offsetOrigin = null) {
	var rectProps = ['width', 'height', 'top', 'right', 'bottom', 'left'];
	var eventProps = ['clientX', 'clientY', 'offsetX', 'offsetY', 'pageX', 'pageY', 'screenX', 'screenY'];
	var _isElement = obj => obj instanceof HTMLElement || obj instanceof HTMLDocument;
	var _rect = {};
	if (el !== window/*window also has screenX and screenY*/ 
	&& _isObject(el) && _any(eventProps, prop => prop in el)) {
		_rect = size !== false ? {width:0, height:0,} : {};
		if (offsetOrigin === window || _isElement(offsetOrigin)) {
			_rect.left = el.clientX;
			_rect.top = el.clientY;
		} else if (offsetOrigin === true) {
			_rect.left = el.offsetX;
			_rect.top = el.offsetY;
		} else if (offsetOrigin !== false) {
			_rect.left = el.pageX;
			_rect.top = el.pageY;
		}
	} else if (el === window) {
		_rect = size !== false ? {width:window.innerWidth, height:window.innerHeight,} : {};
		if (offsetOrigin === window || offsetOrigin === true || _isElement(offsetOrigin)) {
			_rect.left = 0;
			_rect.top = 0;
		} else if (offsetOrigin !== false) {
			_rect.left = window.pageXOffset;// || document.documentElement.scrollLeft;
			_rect.top = window.pageYOffset;// || document.documentElement.scrollTop;
		}
	} else if (_isElement(el)) {
		_rect = el.getBoundingClientRect().toJSON();
		delete _rect.x;
		delete _rect.y;
		delete _rect.right;
		delete _rect.bottom;
		if (size === false) {
			delete _rect.width;
			delete _rect.height;
		}
		if (offsetOrigin === true) {
			offsetOrigin = el.offsetParent;
		} else if (offsetOrigin === false) {
			delete _rect.left;
			delete _rect.top;
		} else if (offsetOrigin !== window && !_isElement(offsetOrigin)) {
			_rect.left += window.pageXOffset;// || document.documentElement.scrollLeft);
			_rect.top += window.pageYOffset;// || document.documentElement.scrollTop);
		}
	} else if (_isObject(el) && _any(rectProps, prop => prop in el)) {
		_rect = _copy(el, rectProps);
	} else {
		throw new Error('Unknown object type!');
	}
	if (_isElement(offsetOrigin)) {
		var contextOffset = rect(offsetOrigin, false, window);
		_rect.left -= contextOffset.left;
		_rect.top -= contextOffset.top;
	}
	return _rect;
};

/**
 * @exports
 */
export default rect;